import pandas as pd
from sklearn.cluster import AgglomerativeClustering

from ._abstract import BaseZigZagLevels, BaseLevelFinder


def _cluster_prices_to_levels(prices, distance, level_selector='mean'):
    clustering = AgglomerativeClustering(distance_threshold=distance, n_clusters=None)
    try:
        clustering.fit(prices.reshape(-1, 1))
    except ValueError:
        return None

    df = pd.DataFrame(data=prices, columns=('price',))
    df['cluster'] = clustering.labels_
    df['peak_count'] = 1

    grouped = df.groupby('cluster').agg(
        {
            'price': level_selector,
            'peak_count': 'sum'
        }
    ).reset_index()

    return grouped.to_dict('records')


class ZigZagClusterLevels(BaseZigZagLevels):

    def _aggregate_prices_to_levels(self, prices, distance):
        return _cluster_prices_to_levels(prices, distance, self._level_selector)


class RawPriceClusterLevels(BaseLevelFinder):
    def __init__(self, merge_distance, merge_percent=None, level_selector='median', use_maximums=True,
                 bars_for_peak=21):

        self._use_max = use_maximums
        self._bars_for_peak = bars_for_peak
        super().__init__(merge_distance, merge_percent, level_selector)

    def _validate_init_args(self):
        super()._validate_init_args()
        if self._bars_for_peak % 2 == 0:
            raise Exception('N bars to define peak should be odd number')

    def _find_potential_level_prices(self, X):
        d = pd.DataFrame(data=X, columns=('price',))
        bars_to_shift = int((self._bars_for_peak - 1) / 2)

        if self._use_max:
            d['F'] = d['price'].rolling(window=self._bars_for_peak).max().shift(-bars_to_shift)
        else:
            d['F'] = d['price'].rolling(window=self._bars_for_peak).min().shift(-bars_to_shift)

        prices = pd.unique(d[d['F'] == d['price']]['price'])

        return prices

    def _aggregate_prices_to_levels(self, prices, distance):
        return _cluster_prices_to_levels(prices, distance, self._level_selector)



from enum import Enum, auto

import numpy as np

from .abstract import BaseScorer


class PointEventType(Enum):
    CUT_BODY = auto()
    CUT_WICK = auto()
    TOUCH_DOWN_HIGHLOW = auto()
    TOUCH_DOWN = auto()
    TOUCH_UP_HIGHLOW = auto()
    TOUCH_UP = auto()


class PointEvent:
    def __init__(self, event_type, timestamp, score_change):
        self.type = event_type
        self.timestamp = timestamp
        self.score_change = score_change


class PointScore:

    def __init__(self, point, score, point_event_list):
        self.point = point
        self.score = score
        self.point_event_list = point_event_list

    def __str__(self):
        return f'| price: {self.point}: score{self.score} |'

    def __repr__(self):
        return str(self)


class TouchScorer(BaseScorer):

    def __init__(self, min_candles_between_body_cuts=5, diff_perc_from_extreme=0.05,
                 min_distance_between_levels=0.1, min_trend_percent=0.5, diff_perc_for_candle_close=0.05,
                 score_for_cut_body=-2, score_for_cut_wick=-1, score_for_touch_high_low=1, score_for_touch_normal=2):

        self.DIFF_PERC_FOR_CANDLE_CLOSE = diff_perc_for_candle_close
        self.MIN_DIFF_FOR_CONSECUTIVE_CUT = min_candles_between_body_cuts
        self.DIFF_PERC_FROM_EXTREME = diff_perc_from_extreme
        self.DIFF_PERC_FOR_INTRASR_DISTANCE = min_distance_between_levels
        self.MIN_PERC_FOR_TREND = min_trend_percent

        self.score_for_cut_body = score_for_cut_body
        self.score_for_cut_wick = score_for_cut_wick
        self.score_for_touch_high_low = score_for_touch_high_low
        self.score_for_touch_normal = score_for_touch_normal

        self._scores = None

    def closeFromExtreme(self, key, min_value, max_value):
        return abs(key - min_value) < (min_value * self.DIFF_PERC_FROM_EXTREME / 100.0) or \
               abs(key - max_value) < (max_value * self.DIFF_PERC_FROM_EXTREME / 100)

    @staticmethod
    def getMin(candles):
        return candles['Low'].min()

    @staticmethod
    def getMax(candles):
        return candles['High'].max()

    def similar(self, key, used):
        for value in used:
            if abs(key - value) <= (self.DIFF_PERC_FOR_INTRASR_DISTANCE * value / 100):
                return True
        return False

    def fit(self, levels, ohlc_df):
        scores = []
        high_low_list = self._get_high_low_list(ohlc_df)

        for i, obj in enumerate(levels):
            if isinstance(obj, float):
                price = obj
            elif isinstance(obj, dict):
                try:
                    price = obj['price']
                except KeyError:
                    raise Exception('`levels` supposed to be a list of floats or list of dicts with `price` key')
            else:
                raise Exception('`levels` supposed to be a list of floats or list of dicts with `price` key')

            score = self._get_level_score(ohlc_df, high_low_list, price)
            scores.append((i, price, score))

        self._scores = scores

    @property
    def scores(self):
        return self._scores

    @staticmethod
    def _get_high_low_list(ohlc_df):
        rolling_lows = ohlc_df['Low'].rolling(window=3).min().shift(-1)
        rolling_highs = ohlc_df['High'].rolling(window=3).min().shift(-1)
        high_low_list = np.where(ohlc_df['Low'] == rolling_lows, True, False)
        high_low_list = np.where(ohlc_df['High'] == rolling_highs, True, high_low_list)

        return high_low_list.tolist()

    def _get_level_score(self, candles, high_low_marks, price):
        events = []
        score = 0.0
        last_cut_pos = -10
        for i in range(len(candles)):
            candle = candles.iloc[i]
            # If the body of the candle cuts through the price, then deduct some score
            if self.cut_body(price, candle) and i - last_cut_pos > self.MIN_DIFF_FOR_CONSECUTIVE_CUT:
                score += self.score_for_cut_body
                last_cut_pos = i
                events.append(PointEvent(PointEventType.CUT_BODY, candle['Datetime'], self.score_for_cut_body))
            # If the wick of the candle cuts through the price, then deduct some score
            elif self.cut_wick(price, candle) and (i - last_cut_pos > self.MIN_DIFF_FOR_CONSECUTIVE_CUT):
                score += self.score_for_cut_wick
                last_cut_pos = i
                events.append(PointEvent(PointEventType.CUT_WICK, candle['Datetime'], self.score_for_cut_body))
            # If the if is close the high of some candle and it was in an uptrend, then add some score to this
            elif self.touch_high(price, candle) and self.in_up_trend(candles, price, i):
                high_low_value = high_low_marks[i]
                # If it is a high, then add some score S1
                if high_low_value:
                    score += self.score_for_touch_high_low
                    events.append(
                        PointEvent(PointEventType.TOUCH_UP_HIGHLOW, candle['Datetime'], self.score_for_touch_high_low))
                # Else add S2. S2 > S1
                else:
                    score += self.score_for_touch_normal
                    events.append(PointEvent(PointEventType.TOUCH_UP, candle['Datetime'], self.score_for_touch_normal))

            # If the if is close the low of some candle and it was in an downtrend, then add some score to this
            elif self.touch_low(price, candle) and self.in_down_trend(candles, price, i):
                high_low_value = high_low_marks[i]
                # If it is a high, then add some score S1
                if high_low_value is not None and not high_low_value:
                    score += self.score_for_touch_high_low
                    events.append(
                        PointEvent(PointEventType.TOUCH_DOWN, candle['Datetime'], self.score_for_touch_high_low))
                # Else add S2. S2 > S1
                else:
                    score += self.score_for_touch_normal
                    events.append(
                        PointEvent(PointEventType.TOUCH_DOWN_HIGHLOW, candle['Datetime'], self.score_for_touch_normal))

        return PointScore(price, score, events)

    def in_down_trend(self, candles, price, start_pos):
        # Either move #MIN_PERC_FOR_TREND in direction of trend, or cut through the price
        pos = start_pos
        while pos >= 0:
            if candles['Low'].iat[pos] < price:
                return False
            if candles['Low'].iat[pos] - price > (price * self.MIN_PERC_FOR_TREND / 100):
                return True
            pos -= 1

        return False

    def in_up_trend(self, candles, price, start_pos):
        pos = start_pos
        while pos >= 0:
            if candles['High'].iat[pos] > price:
                return False
            if price - candles['Low'].iat[pos] > (price * self.MIN_PERC_FOR_TREND / 100):
                return True
            pos -= 1

        return False

    def touch_high(self, price, candle):
        high = candle['High']
        ltp = candle['Close']
        return high <= price and abs(high - price) < ltp * self.DIFF_PERC_FOR_CANDLE_CLOSE / 100

    def touch_low(self, price, candle):
        low = candle['Low']
        ltp = candle['Close']
        return low >= price and abs(low - price) < ltp * self.DIFF_PERC_FOR_CANDLE_CLOSE / 100

    def cut_body(self, point, candle):
        return max(candle['Open'], candle['Close']) > point and min(candle['Open'], candle['Close']) < point

    def cut_wick(self, price, candle):
        return not self.cut_body(price, candle) and candle['High'] > price and candle['Low'] < price



def _plot_levels(where, levels, only_good=False):
    for l in levels:

        if isinstance(l, float):
            where.axhline(y=l, color='black', linestyle='-')
        elif isinstance(l, dict):
            if 'score' in l.keys():
                if only_good and l['score'] < 0:
                    continue
                color = 'red' if l['score'] < 0 else 'blue'
                where.axhline(y=l['price'], color=color, linestyle='-', linewidth=0.2 * abs(l['score']))
            else:
                where.axhline(y=l['price'], color='black', linestyle='-')



import matplotlib.pyplot as plt
import numpy as np
from zigzag import peak_valley_pivots

from ._helpers import _plot_levels


def plot_with_pivots(X, levels, zigzag_percent=1, only_good=False, path=None):
    pivots = peak_valley_pivots(X, zigzag_percent / 100, -zigzag_percent / 100)
    plt.xlim(0, len(X))
    plt.ylim(X.min() * 0.995, X.max() * 1.005)
    plt.plot(np.arange(len(X)), X, 'k-', alpha=0.9)
    plt.plot(np.arange(len(X))[pivots != 0], X[pivots != 0], 'k:', alpha=0.5)

    plt.scatter(np.arange(len(X))[pivots == 1], X[pivots == 1], color='g')
    plt.scatter(np.arange(len(X))[pivots == -1], X[pivots == -1], color='r')

    _plot_levels(plt, levels, only_good)
    if path:
        plt.savefig(path)
    else:
        plt.show()
    plt.close()



import numpy as np
import pandas as pd
from zigzag import peak_valley_pivots

from .exceptions import InvalidParameterException, InvalidArgumentException


class BaseLevelFinder:

    def __init__(self, merge_distance, merge_percent=None, level_selector='median'):

        self._merge_distance = merge_distance
        self._merge_percent = merge_percent

        self._level_selector = level_selector

        self._levels = None
        self._validate_init_args()

    @property
    def levels(self):
        return self._levels

    def _validate_init_args(self):
        pass

    def fit(self, data):
        if isinstance(data, pd.DataFrame):
            X = data['Close'].values
        elif isinstance(data, np.array):
            X = data
        else:
            raise InvalidArgumentException(
                'Only np.array and pd.DataFrame are supported in `fit` method'
            )

        prices = self._find_potential_level_prices(X)
        levels = self._aggregate_prices_to_levels(prices, self._get_distance(X))

        self._levels = levels

    def _find_potential_level_prices(self, X):
        raise NotImplementedError()

    def _get_distance(self, X):
        if self._merge_distance:
            return self._merge_distance

        mean_price = np.mean(X)
        return self._merge_percent * mean_price / 100

    def _aggregate_prices_to_levels(self, pivot_prices, distance):
        raise NotImplementedError()


class BaseZigZagLevels(BaseLevelFinder):

    def __init__(self, peak_percent_delta, merge_distance, merge_percent=None, min_bars_between_peaks=0, peaks='All',
                 level_selector='median'):
        self._peak_percent_delta = peak_percent_delta / 100
        self._min_bars_between_peaks = min_bars_between_peaks
        self._peaks = peaks
        super().__init__(merge_distance, merge_percent, level_selector)

    def _find_potential_level_prices(self, X):
        pivots = peak_valley_pivots(X, self._peak_percent_delta, -self._peak_percent_delta)
        indexes = self._get_pivot_indexes(pivots)
        pivot_prices = X[indexes]

        return pivot_prices

    def _get_pivot_indexes(self, pivots):
        if self._peaks == 'All':
            indexes = np.where(np.abs(pivots) == 1)
        elif self._peaks == 'High':
            indexes = np.where(pivots == 1)
        elif self._peaks == 'Low':
            indexes = np.where(pivots == -1)
        else:
            raise InvalidParameterException(
                'Peaks argument should be one of: `All`, `High`, `Low`'
            )

        return indexes if self._min_bars_between_peaks == 0 else self._filter_by_bars_between(indexes)

    def _filter_by_bars_between(self, indexes):
        indexes = np.sort(indexes).reshape(-1, 1)

        try:
            selected = [indexes[0][0]]
        except IndexError:
            return indexes

        pre_idx = indexes[0][0]
        for i in range(1, len(indexes)):
            if indexes[i][0] - pre_idx < self._min_bars_between_peaks:
                continue
            pre_idx = indexes[i][0]
            selected.append(pre_idx)

        return np.array(selected)



import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
from mpl_finance import candlestick2_ohlc

from ._helpers import _plot_levels


def plot_levels_on_candlestick(df, levels, only_good=False, path=None,
                               formatter=mdates.DateFormatter('%y-%m-%d %H:%M:%S')):
    ohlc = df[['Datetime', 'Open', 'High', 'Low', 'Close']].copy()
    ohlc["Datetime"] = pd.to_datetime(ohlc['Datetime'])
    ohlc["Datetime"] = ohlc["Datetime"].apply(lambda x: mdates.date2num(x))
    f1, ax = plt.subplots(figsize=(10, 5))
    candlestick2_ohlc(ax,
                      closes=ohlc.Close.values,
                      opens=ohlc.Open.values,
                      highs=ohlc.High.values,
                      lows=ohlc.Low.values,
                      colordown='red',
                      colorup='green'
                      )

    _plot_levels(ax, levels, only_good)

    if path:
        plt.savefig(path)
    else:
        plt.show()
    plt.close()



from distutils.util import convert_path
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
metadata = dict()
with open(convert_path('pricelevels/version.py')) as metadata_file:
    exec(metadata_file.read(), metadata)

setup(
    name='pricelevels',

    version=metadata['__version__'],
    zip_safe=False,

    description='Small lib to find support and resistance levels',

    author='Alex Hurko',
    author_email='alex1hurko@gmail.com',

    license='wtfpl',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[
        'scikit-learn==0.21.2',
        'pandas==0.25.0',
        'ZigZag==0.2.2',
        'matplotlib==3.1.1',
    ],
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'pytest'
    ],

    include_package_data=True,

    extras_require={
        'test': ['coverage', 'pytest'],
    },
)
